@@ -5,27 +5,25 @@
 package lsp
 
 import (
+	"bytes"
 	"context"
 	"fmt"
-	"go/token"
 	"os"
-	"os/exec"
 	"path/filepath"
 	"sort"
 	"strings"
 	"testing"
 
 	"github.com/google/go-cmp/cmp"
-	"github.com/google/go-cmp/cmp/cmpopts"
+	"golang.org/x/tools/gopls/internal/bug"
 	"golang.org/x/tools/gopls/internal/lsp/cache"
 	"golang.org/x/tools/gopls/internal/lsp/command"
+	"golang.org/x/tools/gopls/internal/lsp/debug"
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	"golang.org/x/tools/gopls/internal/lsp/source"
 	"golang.org/x/tools/gopls/internal/lsp/tests"
 	"golang.org/x/tools/gopls/internal/lsp/tests/compare"
 	"golang.org/x/tools/gopls/internal/span"
-	"golang.org/x/tools/internal/bug"
-	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/testenv"
 )
 
@@ -33,34 +31,27 @@ func TestMain(m *testing.M) {
 	bug.PanicOnBugs = true
 	testenv.ExitIfSmallMachine()
 
-	// Set the global exporter to nil so that we don't log to stderr. This avoids
-	// a lot of misleading noise in test output.
-	//
-	// TODO(rfindley): investigate whether we can/should capture logs scoped to
-	// individual tests by passing in a context with a local exporter.
-	event.SetExporter(nil)
-
 	os.Exit(m.Run())
 }
 
+// TestLSP runs the marker tests in files beneath testdata/ using
+// implementations of each of the marker operations (e.g. @codelens) that
+// make LSP RPCs (e.g. textDocument/codeLens) to a gopls server.
 func TestLSP(t *testing.T) {
 	tests.RunTests(t, "testdata", true, testLSP)
 }
 
-type runner struct {
-	server      *Server
-	data        *tests.Data
-	diagnostics map[span.URI][]*source.Diagnostic
-	ctx         context.Context
-	normalizers []tests.Normalizer
-	editRecv    chan map[span.URI]string
-}
-
 func testLSP(t *testing.T, datum *tests.Data) {
 	ctx := tests.Context(t)
 
-	cache := cache.New(nil, nil, nil)
-	session := cache.NewSession(ctx)
+	// Setting a debug instance suppresses logging to stderr, but ensures that we
+	// still e.g. convert events into runtime/trace/instrumentation.
+	//
+	// Previously, we called event.SetExporter(nil), which turns off all
+	// instrumentation.
+	ctx = debug.WithInstance(ctx, "", "off")
+
+	session := cache.NewSession(ctx, cache.New(nil), nil)
 	options := source.DefaultOptions().Clone()
 	tests.DefaultOptions(options)
 	session.SetOptions(options)
@@ -70,21 +61,45 @@ func testLSP(t *testing.T, datum *tests.Data) {
 		t.Fatal(err)
 	}
 
-	defer view.Shutdown(ctx)
+	defer session.RemoveView(view)
 
 	// Enable type error analyses for tests.
 	// TODO(golang/go#38212): Delete this once they are enabled by default.
-	tests.EnableAllAnalyzers(view, options)
-	view.SetOptions(ctx, options)
+	tests.EnableAllAnalyzers(options)
+	session.SetViewOptions(ctx, view, options)
 
 	// Enable all inlay hints for tests.
-	tests.EnableAllInlayHints(view, options)
+	tests.EnableAllInlayHints(options)
 
 	// Only run the -modfile specific tests in module mode with Go 1.14 or above.
 	datum.ModfileFlagAvailable = len(snapshot.ModFiles()) > 0 && testenv.Go1Point() >= 14
 	release()
 
+	// Open all files for performance reasons, because gopls only
+	// keeps active packages (those with open files) in memory.
+	//
+	// In practice clients will only send document-oriented requests for open
+	// files.
 	var modifications []source.FileModification
+	for _, module := range datum.Exported.Modules {
+		for name := range module.Files {
+			filename := datum.Exported.File(module.Name, name)
+			if filepath.Ext(filename) != ".go" {
+				continue
+			}
+			content, err := datum.Exported.FileContents(filename)
+			if err != nil {
+				t.Fatal(err)
+			}
+			modifications = append(modifications, source.FileModification{
+				URI:        span.URIFromPath(filename),
+				Action:     source.Open,
+				Version:    -1,
+				Text:       content,
+				LanguageID: "go",
+			})
+		}
+	}
 	for filename, content := range datum.Config.Overlay {
 		if filepath.Ext(filename) != ".go" {
 			continue
@@ -104,13 +119,23 @@ func testLSP(t *testing.T, datum *tests.Data) {
 		data:        datum,
 		ctx:         ctx,
 		normalizers: tests.CollectNormalizers(datum.Exported),
-		editRecv:    make(chan map[span.URI]string, 1),
+		editRecv:    make(chan map[span.URI][]byte, 1),
 	}
 
 	r.server = NewServer(session, testClient{runner: r})
 	tests.Run(t, r, datum)
 }
 
+// runner implements tests.Tests by making LSP RPCs to a gopls server.
+type runner struct {
+	server      *Server
+	data        *tests.Data
+	diagnostics map[span.URI][]*source.Diagnostic
+	ctx         context.Context
+	normalizers []tests.Normalizer
+	editRecv    chan map[span.URI][]byte
+}
+
 // testClient stubs any client functions that may be called by LSP functions.
 type testClient struct {
 	protocol.Client
@@ -145,16 +170,13 @@ func (r *runner) CallHierarchy(t *testing.T, spn span.Span, expectedCalls *tests
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := mapper.Location(spn)
+	loc, err := mapper.SpanLocation(spn)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", spn, err)
 	}
 
 	params := &protocol.CallHierarchyPrepareParams{
-		TextDocumentPositionParams: protocol.TextDocumentPositionParams{
-			TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-			Position:     loc.Range.Start,
-		},
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 
 	items, err := r.server.PrepareCallHierarchy(r.ctx, params)
@@ -219,10 +241,9 @@ func (r *runner) CodeLens(t *testing.T, uri span.URI, want []protocol.CodeLens)
 
 func (r *runner) Diagnostics(t *testing.T, uri span.URI, want []*source.Diagnostic) {
 	// Get the diagnostics for this view if we have not done it before.
-	v := r.server.session.View(r.data.Config.Dir)
+	v := r.server.session.ViewByName(r.data.Config.Dir)
 	r.collectDiagnostics(v)
-	got := append([]*source.Diagnostic(nil), r.diagnostics[uri]...) // copy
-	tests.CompareDiagnostics(t, uri, want, got)
+	tests.CompareDiagnostics(t, uri, want, r.diagnostics[uri])
 }
 
 func (r *runner) FoldingRanges(t *testing.T, spn span.Span) {
@@ -233,43 +254,32 @@ func (r *runner) FoldingRanges(t *testing.T, spn span.Span) {
 	}
 	original := view.Options()
 	modified := original
+	defer r.server.session.SetViewOptions(r.ctx, view, original)
 
-	// Test all folding ranges.
-	modified.LineFoldingOnly = false
-	view, err = view.SetOptions(r.ctx, modified)
-	if err != nil {
-		t.Error(err)
-		return
+	for _, test := range []struct {
+		lineFoldingOnly bool
+		prefix          string
+	}{
+		{false, "foldingRange"},
+		{true, "foldingRange-lineFolding"},
+	} {
+		modified.LineFoldingOnly = test.lineFoldingOnly
+		view, err = r.server.session.SetViewOptions(r.ctx, view, modified)
+		if err != nil {
+			t.Error(err)
+			continue
+		}
+		ranges, err := r.server.FoldingRange(r.ctx, &protocol.FoldingRangeParams{
+			TextDocument: protocol.TextDocumentIdentifier{
+				URI: protocol.URIFromSpanURI(uri),
+			},
+		})
+		if err != nil {
+			t.Error(err)
+			continue
+		}
+		r.foldingRanges(t, test.prefix, uri, ranges)
 	}
-	ranges, err := r.server.FoldingRange(r.ctx, &protocol.FoldingRangeParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-	})
-	if err != nil {
-		t.Error(err)
-		return
-	}
-	r.foldingRanges(t, "foldingRange", uri, ranges)
-
-	// Test folding ranges with lineFoldingOnly = true.
-	modified.LineFoldingOnly = true
-	view, err = view.SetOptions(r.ctx, modified)
-	if err != nil {
-		t.Error(err)
-		return
-	}
-	ranges, err = r.server.FoldingRange(r.ctx, &protocol.FoldingRangeParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-	})
-	if err != nil {
-		t.Error(err)
-		return
-	}
-	r.foldingRanges(t, "foldingRange-lineFolding", uri, ranges)
-	view.SetOptions(r.ctx, original)
 }
 
 func (r *runner) foldingRanges(t *testing.T, prefix string, uri span.URI, ranges []protocol.FoldingRange) {
@@ -355,7 +365,7 @@ func conflict(a, b protocol.FoldingRange) bool {
 		(a.EndLine > b.StartLine || (a.EndLine == b.StartLine && a.EndCharacter > b.StartCharacter))
 }
 
-func foldRanges(m *protocol.ColumnMapper, contents string, ranges []protocol.FoldingRange) (string, error) {
+func foldRanges(m *protocol.Mapper, contents string, ranges []protocol.FoldingRange) (string, error) {
 	foldedText := "<>"
 	res := contents
 	// Apply the edits from the end of the file forward
@@ -363,52 +373,18 @@ func foldRanges(m *protocol.ColumnMapper, contents string, ranges []protocol.Fol
 	// TODO(adonovan): factor to use diff.ApplyEdits, which validates the input.
 	for i := len(ranges) - 1; i >= 0; i-- {
 		r := ranges[i]
-		start, err := m.Point(protocol.Position{r.StartLine, r.StartCharacter})
+		start, end, err := m.RangeOffsets(protocol.Range{
+			Start: protocol.Position{Line: r.StartLine, Character: r.StartCharacter},
+			End:   protocol.Position{Line: r.EndLine, Character: r.EndCharacter},
+		})
 		if err != nil {
 			return "", err
 		}
-		end, err := m.Point(protocol.Position{r.EndLine, r.EndCharacter})
-		if err != nil {
-			return "", err
-		}
-		res = res[:start.Offset()] + foldedText + res[end.Offset():]
+		res = res[:start] + foldedText + res[end:]
 	}
 	return res, nil
 }
 
-func (r *runner) Format(t *testing.T, spn span.Span) {
-	uri := spn.URI()
-	filename := uri.Filename()
-	gofmted := string(r.data.Golden(t, "gofmt", filename, func() ([]byte, error) {
-		cmd := exec.Command("gofmt", filename)
-		out, _ := cmd.Output() // ignore error, sometimes we have intentionally ungofmt-able files
-		return out, nil
-	}))
-
-	edits, err := r.server.Formatting(r.ctx, &protocol.DocumentFormattingParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-	})
-	if err != nil {
-		if gofmted != "" {
-			t.Error(err)
-		}
-		return
-	}
-	m, err := r.data.Mapper(uri)
-	if err != nil {
-		t.Fatal(err)
-	}
-	got, _, err := source.ApplyProtocolEdits(m, edits)
-	if err != nil {
-		t.Error(err)
-	}
-	if diff := compare.Text(gofmted, got); diff != "" {
-		t.Errorf("format failed for %s (-want +got):\n%s", filename, diff)
-	}
-}
-
 func (r *runner) SemanticTokens(t *testing.T, spn span.Span) {
 	uri := spn.URI()
 	filename := uri.Filename()
@@ -442,38 +418,6 @@ func (r *runner) SemanticTokens(t *testing.T, spn span.Span) {
 	}
 }
 
-func (r *runner) Import(t *testing.T, spn span.Span) {
-	uri := spn.URI()
-	filename := uri.Filename()
-	actions, err := r.server.CodeAction(r.ctx, &protocol.CodeActionParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-	})
-	if err != nil {
-		t.Fatal(err)
-	}
-	m, err := r.data.Mapper(uri)
-	if err != nil {
-		t.Fatal(err)
-	}
-	got := string(m.Content)
-	if len(actions) > 0 {
-		res, err := applyTextDocumentEdits(r, actions[0].Edit.DocumentChanges)
-		if err != nil {
-			t.Fatal(err)
-		}
-		got = res[uri]
-	}
-	want := string(r.data.Golden(t, "goimports", filename, func() ([]byte, error) {
-		return []byte(got), nil
-	}))
-
-	if d := compare.Text(want, got); d != "" {
-		t.Errorf("import failed for %s:\n%s", filename, d)
-	}
-}
-
 func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []tests.SuggestedFix, expectedActions int) {
 	uri := spn.URI()
 	view, err := r.server.session.ViewOf(uri)
@@ -485,7 +429,7 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []tests.S
 	if err != nil {
 		t.Fatal(err)
 	}
-	rng, err := m.Range(spn)
+	rng, err := m.SpanRange(spn)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -546,7 +490,7 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []tests.S
 	if !match {
 		t.Fatalf("unexpected kind for code action %s, got %v, want one of %v", action.Title, action.Kind, codeActionKinds)
 	}
-	var res map[span.URI]string
+	var res map[span.URI][]byte
 	if cmd := action.Command; cmd != nil {
 		_, err := r.server.ExecuteCommand(r.ctx, &protocol.ExecuteCommandParams{
 			Command:   action.Command.Command,
@@ -563,63 +507,11 @@ func (r *runner) SuggestedFix(t *testing.T, spn span.Span, actionKinds []tests.S
 		}
 	}
 	for u, got := range res {
-		want := string(r.data.Golden(t, "suggestedfix_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
-			return []byte(got), nil
-		}))
-		if want != got {
-			t.Errorf("suggested fixes failed for %s:\n%s", u.Filename(), compare.Text(want, got))
-		}
-	}
-}
-
-func (r *runner) FunctionExtraction(t *testing.T, start span.Span, end span.Span) {
-	uri := start.URI()
-	m, err := r.data.Mapper(uri)
-	if err != nil {
-		t.Fatal(err)
-	}
-	spn := span.New(start.URI(), start.Start(), end.End())
-	rng, err := m.Range(spn)
-	if err != nil {
-		t.Fatal(err)
-	}
-	actionsRaw, err := r.server.CodeAction(r.ctx, &protocol.CodeActionParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-		Range: rng,
-		Context: protocol.CodeActionContext{
-			Only: []protocol.CodeActionKind{"refactor.extract"},
-		},
-	})
-	if err != nil {
-		t.Fatal(err)
-	}
-	var actions []protocol.CodeAction
-	for _, action := range actionsRaw {
-		if action.Command.Title == "Extract function" {
-			actions = append(actions, action)
-		}
-	}
-	// Hack: We assume that we only get one code action per range.
-	// TODO(rstambler): Support multiple code actions per test.
-	if len(actions) == 0 || len(actions) > 1 {
-		t.Fatalf("unexpected number of code actions, want 1, got %v", len(actions))
-	}
-	_, err = r.server.ExecuteCommand(r.ctx, &protocol.ExecuteCommandParams{
-		Command:   actions[0].Command.Command,
-		Arguments: actions[0].Command.Arguments,
-	})
-	if err != nil {
-		t.Fatal(err)
-	}
-	res := <-r.editRecv
-	for u, got := range res {
-		want := string(r.data.Golden(t, "functionextraction_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
-			return []byte(got), nil
-		}))
-		if want != got {
-			t.Errorf("function extraction failed for %s:\n%s", u.Filename(), compare.Text(want, got))
+		want := r.data.Golden(t, "suggestedfix_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
+			return got, nil
+		})
+		if diff := compare.Bytes(want, got); diff != "" {
+			t.Errorf("suggested fixes failed for %s:\n%s", u.Filename(), diff)
 		}
 	}
 }
@@ -631,7 +523,7 @@ func (r *runner) MethodExtraction(t *testing.T, start span.Span, end span.Span)
 		t.Fatal(err)
 	}
 	spn := span.New(start.URI(), start.Start(), end.End())
-	rng, err := m.Range(spn)
+	rng, err := m.SpanRange(spn)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -667,40 +559,39 @@ func (r *runner) MethodExtraction(t *testing.T, start span.Span, end span.Span)
 	}
 	res := <-r.editRecv
 	for u, got := range res {
-		want := string(r.data.Golden(t, "methodextraction_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
-			return []byte(got), nil
-		}))
-		if want != got {
-			t.Errorf("method extraction failed for %s:\n%s", u.Filename(), compare.Text(want, got))
+		want := r.data.Golden(t, "methodextraction_"+tests.SpanName(spn), u.Filename(), func() ([]byte, error) {
+			return got, nil
+		})
+		if diff := compare.Bytes(want, got); diff != "" {
+			t.Errorf("method extraction failed for %s:\n%s", u.Filename(), diff)
 		}
 	}
 }
 
-func (r *runner) Definition(t *testing.T, spn span.Span, d tests.Definition) {
+// TODO(rfindley): This handler needs more work. The output is still a bit hard
+// to read (range diffs do not format nicely), and it is too entangled with hover.
+func (r *runner) Definition(t *testing.T, _ span.Span, d tests.Definition) {
 	sm, err := r.data.Mapper(d.Src.URI())
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := sm.Location(d.Src)
+	loc, err := sm.SpanLocation(d.Src)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", d.Src, err)
 	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
-	}
-	var locs []protocol.Location
+	tdpp := protocol.LocationTextDocumentPositionParams(loc)
+	var got []protocol.Location
 	var hover *protocol.Hover
 	if d.IsType {
 		params := &protocol.TypeDefinitionParams{
 			TextDocumentPositionParams: tdpp,
 		}
-		locs, err = r.server.TypeDefinition(r.ctx, params)
+		got, err = r.server.TypeDefinition(r.ctx, params)
 	} else {
 		params := &protocol.DefinitionParams{
 			TextDocumentPositionParams: tdpp,
 		}
-		locs, err = r.server.Definition(r.ctx, params)
+		got, err = r.server.Definition(r.ctx, params)
 		if err != nil {
 			t.Fatalf("failed for %v: %+v", d.Src, err)
 		}
@@ -712,31 +603,41 @@ func (r *runner) Definition(t *testing.T, spn span.Span, d tests.Definition) {
 	if err != nil {
 		t.Fatalf("failed for %v: %v", d.Src, err)
 	}
-	if len(locs) != 1 {
-		t.Errorf("got %d locations for definition, expected 1", len(locs))
+	dm, err := r.data.Mapper(d.Def.URI())
+	if err != nil {
+		t.Fatal(err)
+	}
+	def, err := dm.SpanLocation(d.Def)
+	if err != nil {
+		t.Fatal(err)
+	}
+	if !d.OnlyHover {
+		want := []protocol.Location{def}
+		if diff := cmp.Diff(want, got); diff != "" {
+			t.Fatalf("Definition(%s) mismatch (-want +got):\n%s", d.Src, diff)
+		}
 	}
 	didSomething := false
 	if hover != nil {
 		didSomething = true
 		tag := fmt.Sprintf("%s-hoverdef", d.Name)
-		expectHover := string(r.data.Golden(t, tag, d.Src.URI().Filename(), func() ([]byte, error) {
+		want := string(r.data.Golden(t, tag, d.Src.URI().Filename(), func() ([]byte, error) {
 			return []byte(hover.Contents.Value), nil
 		}))
-		got := tests.StripSubscripts(hover.Contents.Value)
-		expectHover = tests.StripSubscripts(expectHover)
-		if got != expectHover {
-			tests.CheckSameMarkdown(t, got, expectHover)
+		got := hover.Contents.Value
+		if diff := tests.DiffMarkdown(want, got); diff != "" {
+			t.Errorf("%s: markdown mismatch:\n%s", d.Src, diff)
 		}
 	}
 	if !d.OnlyHover {
 		didSomething = true
-		locURI := locs[0].URI.SpanURI()
+		locURI := got[0].URI.SpanURI()
 		lm, err := r.data.Mapper(locURI)
 		if err != nil {
 			t.Fatal(err)
 		}
-		if def, err := lm.Span(locs[0]); err != nil {
-			t.Fatalf("failed for %v: %v", locs[0], err)
+		if def, err := lm.LocationSpan(got[0]); err != nil {
+			t.Fatalf("failed for %v: %v", got[0], err)
 		} else if def != d.Def {
 			t.Errorf("for %v got %v want %v", d.Src, def, d.Def)
 		}
@@ -746,189 +647,48 @@ func (r *runner) Definition(t *testing.T, spn span.Span, d tests.Definition) {
 	}
 }
 
-func (r *runner) Implementation(t *testing.T, spn span.Span, impls []span.Span) {
-	sm, err := r.data.Mapper(spn.URI())
-	if err != nil {
-		t.Fatal(err)
-	}
-	loc, err := sm.Location(spn)
-	if err != nil {
-		t.Fatalf("failed for %v: %v", spn, err)
-	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
-	}
-	var locs []protocol.Location
-	params := &protocol.ImplementationParams{
-		TextDocumentPositionParams: tdpp,
-	}
-	locs, err = r.server.Implementation(r.ctx, params)
-	if err != nil {
-		t.Fatalf("failed for %v: %v", spn, err)
-	}
-	if len(locs) != len(impls) {
-		t.Fatalf("got %d locations for implementation, expected %d", len(locs), len(impls))
-	}
-
-	var results []span.Span
-	for i := range locs {
-		locURI := locs[i].URI.SpanURI()
-		lm, err := r.data.Mapper(locURI)
-		if err != nil {
-			t.Fatal(err)
-		}
-		imp, err := lm.Span(locs[i])
-		if err != nil {
-			t.Fatalf("failed for %v: %v", locs[i], err)
-		}
-		results = append(results, imp)
-	}
-	// Sort results and expected to make tests deterministic.
-	sort.SliceStable(results, func(i, j int) bool {
-		return span.Compare(results[i], results[j]) == -1
-	})
-	sort.SliceStable(impls, func(i, j int) bool {
-		return span.Compare(impls[i], impls[j]) == -1
-	})
-	for i := range results {
-		if results[i] != impls[i] {
-			t.Errorf("for %dth implementation of %v got %v want %v", i, spn, results[i], impls[i])
-		}
-	}
-}
-
-func (r *runner) Highlight(t *testing.T, src span.Span, locations []span.Span) {
+func (r *runner) Highlight(t *testing.T, src span.Span, spans []span.Span) {
 	m, err := r.data.Mapper(src.URI())
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := m.Location(src)
+	loc, err := m.SpanLocation(src)
 	if err != nil {
-		t.Fatalf("failed for %v: %v", locations[0], err)
-	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
+		t.Fatal(err)
 	}
 	params := &protocol.DocumentHighlightParams{
-		TextDocumentPositionParams: tdpp,
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 	highlights, err := r.server.DocumentHighlight(r.ctx, params)
 	if err != nil {
-		t.Fatal(err)
+		t.Fatalf("DocumentHighlight(%v) failed: %v", params, err)
 	}
-	if len(highlights) != len(locations) {
-		t.Fatalf("got %d highlights for highlight at %v:%v:%v, expected %d", len(highlights), src.URI().Filename(), src.Start().Line(), src.Start().Column(), len(locations))
+	var got []protocol.Range
+	for _, h := range highlights {
+		got = append(got, h.Range)
 	}
-	// Check to make sure highlights have a valid range.
-	var results []span.Span
-	for i := range highlights {
-		h, err := m.RangeSpan(highlights[i].Range)
+
+	var want []protocol.Range
+	for _, s := range spans {
+		rng, err := m.SpanRange(s)
 		if err != nil {
-			t.Fatalf("failed for %v: %v", highlights[i], err)
+			t.Fatalf("Mapper.SpanRange(%v) failed: %v", s, err)
 		}
-		results = append(results, h)
+		want = append(want, rng)
 	}
-	// Sort results to make tests deterministic since DocumentHighlight uses a map.
-	sort.SliceStable(results, func(i, j int) bool {
-		return span.Compare(results[i], results[j]) == -1
-	})
-	// Check to make sure all the expected highlights are found.
-	for i := range results {
-		if results[i] != locations[i] {
-			t.Errorf("want %v, got %v\n", locations[i], results[i])
-		}
-	}
-}
 
-func (r *runner) Hover(t *testing.T, src span.Span, text string) {
-	m, err := r.data.Mapper(src.URI())
-	if err != nil {
-		t.Fatal(err)
-	}
-	loc, err := m.Location(src)
-	if err != nil {
-		t.Fatalf("failed for %v", err)
-	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
-	}
-	params := &protocol.HoverParams{
-		TextDocumentPositionParams: tdpp,
-	}
-	hover, err := r.server.Hover(r.ctx, params)
-	if err != nil {
-		t.Fatal(err)
-	}
-	if text == "" {
-		if hover != nil {
-			t.Errorf("want nil, got %v\n", hover)
-		}
-	} else {
-		if hover == nil {
-			t.Fatalf("want hover result to include %s, but got nil", text)
-		}
-		if got := hover.Contents.Value; got != text {
-			t.Errorf("want %v, got %v\n", text, got)
-		}
-		if want, got := loc.Range, hover.Range; want != got {
-			t.Errorf("want range %v, got %v instead", want, got)
-		}
-	}
-}
-
-func (r *runner) References(t *testing.T, src span.Span, itemList []span.Span) {
-	sm, err := r.data.Mapper(src.URI())
-	if err != nil {
-		t.Fatal(err)
-	}
-	loc, err := sm.Location(src)
-	if err != nil {
-		t.Fatalf("failed for %v: %v", src, err)
-	}
-	for _, includeDeclaration := range []bool{true, false} {
-		t.Run(fmt.Sprintf("refs-declaration-%v", includeDeclaration), func(t *testing.T) {
-			want := make(map[protocol.Location]bool)
-			for i, pos := range itemList {
-				// We don't want the first result if we aren't including the declaration.
-				if i == 0 && !includeDeclaration {
-					continue
-				}
-				m, err := r.data.Mapper(pos.URI())
-				if err != nil {
-					t.Fatal(err)
-				}
-				loc, err := m.Location(pos)
-				if err != nil {
-					t.Fatalf("failed for %v: %v", src, err)
-				}
-				want[loc] = true
-			}
-			params := &protocol.ReferenceParams{
-				TextDocumentPositionParams: protocol.TextDocumentPositionParams{
-					TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-					Position:     loc.Range.Start,
-				},
-				Context: protocol.ReferenceContext{
-					IncludeDeclaration: includeDeclaration,
-				},
-			}
-			got, err := r.server.References(r.ctx, params)
-			if err != nil {
-				t.Fatalf("failed for %v: %v", src, err)
-			}
-			if len(got) != len(want) {
-				t.Errorf("references failed: different lengths got %v want %v", len(got), len(want))
-			}
-			for _, loc := range got {
-				if !want[loc] {
-					t.Errorf("references failed: incorrect references got %v want %v", loc, want)
-				}
-			}
+	sortRanges := func(s []protocol.Range) {
+		sort.Slice(s, func(i, j int) bool {
+			return protocol.CompareRange(s[i], s[j]) < 0
 		})
 	}
+
+	sortRanges(got)
+	sortRanges(want)
+
+	if diff := cmp.Diff(want, got); diff != "" {
+		t.Errorf("DocumentHighlight(%v) mismatch (-want +got):\n%s", src, diff)
+	}
 }
 
 func (r *runner) InlayHints(t *testing.T, spn span.Span) {
@@ -956,7 +716,7 @@ func (r *runner) InlayHints(t *testing.T, spn span.Span) {
 			paddingRight = " "
 		}
 		edits[i] = protocol.TextEdit{
-			Range:   protocol.Range{Start: *hint.Position, End: *hint.Position},
+			Range:   protocol.Range{Start: hint.Position, End: hint.Position},
 			NewText: fmt.Sprintf("<%s%s%s>", paddingLeft, hint.Label[0].Value, paddingRight),
 		}
 	}
@@ -970,12 +730,12 @@ func (r *runner) InlayHints(t *testing.T, spn span.Span) {
 		t.Error(err)
 	}
 
-	withinlayHints := string(r.data.Golden(t, "inlayHint", filename, func() ([]byte, error) {
-		return []byte(got), nil
-	}))
+	withinlayHints := r.data.Golden(t, "inlayHint", filename, func() ([]byte, error) {
+		return got, nil
+	})
 
-	if withinlayHints != got {
-		t.Errorf("inlay hints failed for %s, expected:\n%v\ngot:\n%v", filename, withinlayHints, got)
+	if !bytes.Equal(withinlayHints, got) {
+		t.Errorf("inlay hints failed for %s, expected:\n%s\ngot:\n%s", filename, withinlayHints, got)
 	}
 }
 
@@ -988,24 +748,22 @@ func (r *runner) Rename(t *testing.T, spn span.Span, newText string) {
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := sm.Location(spn)
+	loc, err := sm.SpanLocation(spn)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", spn, err)
 	}
 
 	wedit, err := r.server.Rename(r.ctx, &protocol.RenameParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-		Position: loc.Range.Start,
-		NewName:  newText,
+		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
+		Position:     loc.Range.Start,
+		NewName:      newText,
 	})
 	if err != nil {
 		renamed := string(r.data.Golden(t, tag, filename, func() ([]byte, error) {
 			return []byte(err.Error()), nil
 		}))
 		if err.Error() != renamed {
-			t.Errorf("rename failed for %s, expected:\n%v\ngot:\n%v\n", newText, renamed, err)
+			t.Errorf("%s: rename failed for %s, expected:\n%v\ngot:\n%v\n", spn, newText, renamed, err)
 		}
 		return
 	}
@@ -1019,23 +777,26 @@ func (r *runner) Rename(t *testing.T, spn span.Span, newText string) {
 	}
 	sort.Strings(orderedURIs)
 
-	var got string
+	// Print the name and content of each modified file,
+	// concatenated, and compare against the golden.
+	var buf bytes.Buffer
 	for i := 0; i < len(res); i++ {
 		if i != 0 {
-			got += "\n"
+			buf.WriteByte('\n')
 		}
 		uri := span.URIFromURI(orderedURIs[i])
 		if len(res) > 1 {
-			got += filepath.Base(uri.Filename()) + ":\n"
+			buf.WriteString(filepath.Base(uri.Filename()))
+			buf.WriteString(":\n")
 		}
-		val := res[uri]
-		got += val
+		buf.Write(res[uri])
 	}
-	want := string(r.data.Golden(t, tag, filename, func() ([]byte, error) {
-		return []byte(got), nil
-	}))
-	if want != got {
-		t.Errorf("rename failed for %s:\n%s", newText, compare.Text(want, got))
+	got := buf.Bytes()
+	want := r.data.Golden(t, tag, filename, func() ([]byte, error) {
+		return got, nil
+	})
+	if diff := compare.Bytes(want, got); diff != "" {
+		t.Errorf("rename failed for %s:\n%s", newText, diff)
 	}
 }
 
@@ -1044,16 +805,12 @@ func (r *runner) PrepareRename(t *testing.T, src span.Span, want *source.Prepare
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := m.Location(src)
+	loc, err := m.SpanLocation(src)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", src, err)
 	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{URI: loc.URI},
-		Position:     loc.Range.Start,
-	}
 	params := &protocol.PrepareRenameParams{
-		TextDocumentPositionParams: tdpp,
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 	got, err := r.server.PrepareRename(context.Background(), params)
 	if err != nil {
@@ -1088,16 +845,16 @@ func (r *runner) PrepareRename(t *testing.T, src span.Span, want *source.Prepare
 	}
 }
 
-func applyTextDocumentEdits(r *runner, edits []protocol.DocumentChanges) (map[span.URI]string, error) {
-	res := map[span.URI]string{}
+func applyTextDocumentEdits(r *runner, edits []protocol.DocumentChanges) (map[span.URI][]byte, error) {
+	res := make(map[span.URI][]byte)
 	for _, docEdits := range edits {
 		if docEdits.TextDocumentEdit != nil {
 			uri := docEdits.TextDocumentEdit.TextDocument.URI.SpanURI()
-			var m *protocol.ColumnMapper
+			var m *protocol.Mapper
 			// If we have already edited this file, we use the edited version (rather than the
 			// file in its original state) so that we preserve our initial changes.
 			if content, ok := res[uri]; ok {
-				m = protocol.NewColumnMapper(uri, []byte(content))
+				m = protocol.NewMapper(uri, content)
 			} else {
 				var err error
 				if m, err = r.data.Mapper(uri); err != nil {
@@ -1114,89 +871,17 @@ func applyTextDocumentEdits(r *runner, edits []protocol.DocumentChanges) (map[sp
 	return res, nil
 }
 
-func (r *runner) Symbols(t *testing.T, uri span.URI, expectedSymbols []protocol.DocumentSymbol) {
-	params := &protocol.DocumentSymbolParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(uri),
-		},
-	}
-	got, err := r.server.DocumentSymbol(r.ctx, params)
-	if err != nil {
-		t.Fatal(err)
-	}
-
-	symbols := make([]protocol.DocumentSymbol, len(got))
-	for i, s := range got {
-		s, ok := s.(protocol.DocumentSymbol)
-		if !ok {
-			t.Fatalf("%v: wanted []DocumentSymbols but got %v", uri, got)
-		}
-		symbols[i] = s
-	}
-
-	// Sort by position to make it easier to find errors.
-	sortSymbols := func(s []protocol.DocumentSymbol) {
-		sort.Slice(s, func(i, j int) bool {
-			return protocol.CompareRange(s[i].SelectionRange, s[j].SelectionRange) < 0
-		})
-	}
-	sortSymbols(expectedSymbols)
-	sortSymbols(symbols)
-
-	// Ignore 'Range' here as it is difficult (impossible?) to express
-	// multi-line ranges in the packagestest framework.
-	ignoreRange := cmpopts.IgnoreFields(protocol.DocumentSymbol{}, "Range")
-	if diff := cmp.Diff(expectedSymbols, symbols, ignoreRange); diff != "" {
-		t.Errorf("mismatching symbols (-want +got)\n%s", diff)
-	}
-}
-
-func (r *runner) WorkspaceSymbols(t *testing.T, uri span.URI, query string, typ tests.WorkspaceSymbolsTestType) {
-	matcher := tests.WorkspaceSymbolsTestTypeToMatcher(typ)
-
-	original := r.server.session.Options()
-	modified := original
-	modified.SymbolMatcher = matcher
-	r.server.session.SetOptions(modified)
-	defer r.server.session.SetOptions(original)
-
-	params := &protocol.WorkspaceSymbolParams{
-		Query: query,
-	}
-	gotSymbols, err := r.server.Symbol(r.ctx, params)
-	if err != nil {
-		t.Fatal(err)
-	}
-	got, err := tests.WorkspaceSymbolsString(r.ctx, r.data, uri, gotSymbols)
-	if err != nil {
-		t.Fatal(err)
-	}
-	got = filepath.ToSlash(tests.Normalize(got, r.normalizers))
-	want := string(r.data.Golden(t, fmt.Sprintf("workspace_symbol-%s-%s", strings.ToLower(string(matcher)), query), uri.Filename(), func() ([]byte, error) {
-		return []byte(got), nil
-	}))
-	if diff := compare.Text(want, got); diff != "" {
-		t.Error(diff)
-	}
-}
-
 func (r *runner) SignatureHelp(t *testing.T, spn span.Span, want *protocol.SignatureHelp) {
 	m, err := r.data.Mapper(spn.URI())
 	if err != nil {
 		t.Fatal(err)
 	}
-	loc, err := m.Location(spn)
+	loc, err := m.SpanLocation(spn)
 	if err != nil {
 		t.Fatalf("failed for %v: %v", loc, err)
 	}
-	tdpp := protocol.TextDocumentPositionParams{
-		TextDocument: protocol.TextDocumentIdentifier{
-			URI: protocol.URIFromSpanURI(spn.URI()),
-		},
-		Position: loc.Range.Start,
-	}
 	params := &protocol.SignatureHelpParams{
-		TextDocumentPositionParams: tdpp,
+		TextDocumentPositionParams: protocol.LocationTextDocumentPositionParams(loc),
 	}
 	got, err := r.server.SignatureHelp(r.ctx, params)
 	if err != nil {
@@ -1284,61 +969,86 @@ func (r *runner) AddImport(t *testing.T, uri span.URI, expectedImport string) {
 	if want == nil {
 		t.Fatalf("golden file %q not found", uri.Filename())
 	}
-	if diff := compare.Text(got, string(want)); diff != "" {
+	if diff := compare.Bytes(want, got); diff != "" {
 		t.Errorf("%s mismatch\n%s", command.AddImport, diff)
 	}
 }
 
-func TestBytesOffset(t *testing.T) {
-	tests := []struct {
-		text string
-		pos  protocol.Position
-		want int
-	}{
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 0}, want: 0},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 1}, want: 1},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 2}, want: 1},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 3}, want: 5},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 4}, want: 6},
-		{text: `a𐐀b`, pos: protocol.Position{Line: 0, Character: 5}, want: -1},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 0, Character: 3}, want: 3},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 0, Character: 4}, want: 3},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 1, Character: 0}, want: 4},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 1, Character: 3}, want: 7},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 1, Character: 4}, want: 7},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 2, Character: 0}, want: 8},
-		{text: "aaa\nbbb\n", pos: protocol.Position{Line: 2, Character: 1}, want: -1},
-		{text: "aaa\nbbb\n\n", pos: protocol.Position{Line: 2, Character: 0}, want: 8},
+func (r *runner) SelectionRanges(t *testing.T, spn span.Span) {
+	uri := spn.URI()
+	sm, err := r.data.Mapper(uri)
+	if err != nil {
+		t.Fatal(err)
+	}
+	loc, err := sm.SpanLocation(spn)
+	if err != nil {
+		t.Error(err)
 	}
 
-	for i, test := range tests {
-		fname := fmt.Sprintf("test %d", i)
-		fset := token.NewFileSet()
-		f := fset.AddFile(fname, -1, len(test.text))
-		f.SetLinesForContent([]byte(test.text))
-		uri := span.URIFromPath(fname)
-		mapper := protocol.NewColumnMapper(uri, []byte(test.text))
-		got, err := mapper.Point(test.pos)
-		if err != nil && test.want != -1 {
-			t.Errorf("unexpected error: %v", err)
-		}
-		if err == nil && got.Offset() != test.want {
-			t.Errorf("want %d for %q(Line:%d,Character:%d), but got %d", test.want, test.text, int(test.pos.Line), int(test.pos.Character), got.Offset())
+	ranges, err := r.server.selectionRange(r.ctx, &protocol.SelectionRangeParams{
+		TextDocument: protocol.TextDocumentIdentifier{
+			URI: protocol.URIFromSpanURI(uri),
+		},
+		Positions: []protocol.Position{loc.Range.Start},
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	sb := &strings.Builder{}
+	for i, path := range ranges {
+		fmt.Fprintf(sb, "Ranges %d: ", i)
+		rng := path
+		for {
+			s, e, err := sm.RangeOffsets(rng.Range)
+			if err != nil {
+				t.Error(err)
+			}
+
+			var snippet string
+			if e-s < 30 {
+				snippet = string(sm.Content[s:e])
+			} else {
+				snippet = string(sm.Content[s:s+15]) + "..." + string(sm.Content[e-15:e])
+			}
+
+			fmt.Fprintf(sb, "\n\t%v %q", rng.Range, strings.ReplaceAll(snippet, "\n", "\\n"))
+
+			if rng.Parent == nil {
+				break
+			}
+			rng = *rng.Parent
 		}
+		sb.WriteRune('\n')
+	}
+	got := sb.String()
+
+	testName := "selectionrange_" + tests.SpanName(spn)
+	want := r.data.Golden(t, testName, uri.Filename(), func() ([]byte, error) {
+		return []byte(got), nil
+	})
+	if want == nil {
+		t.Fatalf("golden file %q not found", uri.Filename())
+	}
+	if diff := compare.Text(got, string(want)); diff != "" {
+		t.Errorf("%s mismatch\n%s", testName, diff)
 	}
 }
 
-func (r *runner) collectDiagnostics(view source.View) {
+func (r *runner) collectDiagnostics(view *cache.View) {
 	if r.diagnostics != nil {
 		return
 	}
 	r.diagnostics = make(map[span.URI][]*source.Diagnostic)
 
-	snapshot, release := view.Snapshot(r.ctx)
+	snapshot, release, err := view.Snapshot()
+	if err != nil {
+		panic(err)
+	}
 	defer release()
 
 	// Always run diagnostics with analysis.
-	r.server.diagnose(r.ctx, snapshot, true)
+	r.server.diagnose(r.ctx, snapshot, analyzeEverything)
 	for uri, reports := range r.server.diagnostics {
 		for _, report := range reports.reports {
 			for _, d := range report.diags {
